"""
DNARecords helper utilities.
"""


class DNARecordsUtils:
    """Utility class to provide common functionalities used in other modules.
    """
    from typing import Dict
    from types import ModuleType
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from pyspark.sql import SparkSession

    @staticmethod
    def spark_session() -> 'SparkSession':
        """Gets the current spark session or builds a new one if none.

        Ensures sparktfrecord libraries are available in the session.

        :return: a spark session with sparktfrecord libraries available.
        :rtype: SparkSession
        """
        from pyspark.sql import SparkSession
        return SparkSession.builder \
            .config('spark.jars.packages', 'com.linkedin.sparktfrecord:spark-tfrecord_2.12:0.3.4') \
            .getOrCreate()

    @staticmethod
    def init_hail() -> ModuleType:
        """Initializes Hail ensuring sparktfrecord libraries are available in the session.
        :return: the hail module (with Hail initialized)
        :rtype: ModuleType
        """
        import hail as hl

        conf = {'spark.jars.packages': 'com.linkedin.sparktfrecord:spark-tfrecord_2.12:0.3.4'}
        hl.init(idempotent=True, log='/tmp/hail.log', spark_conf=conf)
        return hl

    @staticmethod
    def dnarecords_tree(dnarecords_path) -> Dict[str, str]:
        """DNARecords directory structure.

        Gets a dictionary with the full structure of a DNARecords dataset given a root path.

        .. code-block:: text

            swrec -> <dnarecords_path>/data/swrec (sample wise dna tfrecords)
            vwrec -> <dnarecords_path>/data/vwrec (variant wise dna tfrecords)
            swpar -> <dnarecords_path>/data/swpar (sample wise dna parquet files)
            vwpar -> <dnarecords_path>/data/vwpar (variant wise dna parquet files)
            skeys -> <dnarecords_path>/meta/skeys (sample wise key mapping)
            vkeys -> <dnarecords_path>/meta/vkeys (variant wise key mapping)
            swpfs -> <dnarecords_path>/meta/swpfs (sample wise parquet files index)
            vwpfs -> <dnarecords_path>/meta/vwpfs (variant wise parquet files index)
            swrfs -> <dnarecords_path>/meta/swrfs (sample wise tfrecords index)
            vwrfs -> <dnarecords_path>/meta/vwrfs (variant wise tfrecords index)
            swpsc -> <dnarecords_path>/meta/swpsc (sample wise parquet schema)
            vwpsc -> <dnarecords_path>/meta/vwpsc (variant wise parquet schema)
            swrsc -> <dnarecords_path>/meta/swrsc (sample wise tfrecord schema)
            vwrsc -> <dnarecords_path>/meta/vwrsc (variant wise tfrecord schema)

        :return: a dictionary with the structure of the DNARecords dataset.
        :rtype: Dict[str,str]
        """
        return {'swrec': f'{dnarecords_path}/data/swrec',
                'vwrec': f'{dnarecords_path}/data/vwrec',
                'swpar': f'{dnarecords_path}/data/swpar',
                'vwpar': f'{dnarecords_path}/data/vwpar',
                'skeys': f'{dnarecords_path}/meta/skeys',
                'vkeys': f'{dnarecords_path}/meta/vkeys',
                'swpfs': f'{dnarecords_path}/meta/swpfs',
                'vwpfs': f'{dnarecords_path}/meta/vwpfs',
                'swrfs': f'{dnarecords_path}/meta/swrfs',
                'vwrfs': f'{dnarecords_path}/meta/vwrfs',
                'swpsc': f'{dnarecords_path}/meta/swpsc',
                'vwpsc': f'{dnarecords_path}/meta/vwpsc',
                'swrsc': f'{dnarecords_path}/meta/swrsc',
                'vwrsc': f'{dnarecords_path}/meta/vwrsc'}



"""
DNARecords available readers.
"""


class DNARecordsReader:
    """DNARecords Tensorflow reader. **Sample and variant wise**.

    Genomics data ML ready for frameworks like Tensorflow or Pytorch.

    * Consume the data in a **variant wise fashion (common GWAS analysis)**.
    * Or consume the data in a **sample wise fashion (Deep Learning models)**.
    * **Tested on UKBB**.

    Example
    --------

    .. code-block:: python

        import dnarecords as dr
        import tensorflow as tf


        output = '/tmp/dnarecords/output'
        reader = dr.reader.DNARecordsReader('/tmp/dnarecords/output')

        swds = reader.sample_wise_dataset()
        tf.print(next(iter(swds)))

        vwds = reader.variant_wise_dataset()
        tf.print(next(iter(vwds)))


    .. code-block:: text

        {'key': 191,
         'chr1': 'SparseTensor(indices=[[0] [1] [2] ... [924] [925] [926]],
                               values=[0.200760081 0.200760037 0.200760067 ... 0.0019912892 1.96934652 0.00396528561],
                               shape=[909])',
         'chr10': 'SparseTensor(indices=[[124] [125] [126] ... [665] [666] [667]],
                                values=[1.01560163 0.0306534301 1.99800873 ... 0.999999881 1.01956224 0.111815773],
                                shape=[532])',
          ... }
         ...

        {'key': 3764,
         'tensor': 'SparseTensor(indices=[[0] [1] [2] ... [281] [282] [283]],
                                 values=[0.111815773 0.015601662 0.00788068399 ... 0.0593509413 0.000500936178],
                                 shape=[10880])'}
        ...

    :param dnarecords_path: root path to your DNARecords created with :obj:`.DNARecordsWriter.write`
    :param gzip: whether your tfrecords are gzipped or not. Default: True.

    See Also
    --------
    :obj:`.DNARecordsWriter.write`
    """

    from typing import Dict, List
    from pandas import DataFrame
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from tensorflow.python.data import Dataset

    def __init__(self, dnarecords_path, gzip=True):
        self._dnarecords_path = dnarecords_path
        self._gzip = gzip

    @staticmethod
    def _types_dict():
        import tensorflow as tf
        return {'long': tf.int64,
                'double': tf.float32}

    @staticmethod
    def _pandas_safe_read_parquet(path, columns, taste):
        import pandas as pd
        import tensorflow as tf

        files = tf.io.gfile.glob(f'{path}/*.parquet')
        if files:
            if columns:
                if taste:
                    return pd.read_parquet(files[0], columns=columns)
                else:
                    return pd.concat(pd.read_parquet(f, columns=columns) for f in files)
            if not columns:
                if taste:
                    return pd.read_parquet(files[0])
                else:
                    return pd.concat(pd.read_parquet(f) for f in files)
        return None

    @staticmethod
    def _pandas_safe_read_json(path):
        import pandas as pd
        import tensorflow as tf

        files = tf.io.gfile.glob(f'{path}/*.json')
        if files:
            return pd.concat(pd.read_json(f) for f in files)
        return None

    def metadata(self, vkeys_columns: List[str] = None, skeys_columns: List[str] = None, taste: bool = False) -> Dict[str, DataFrame]:
        """Gets the metadata associated to the DNARecords dataset as a dictionary of names to pandas DataFrames.

        :rtype: Dict[str, DataFrame].
        :return: the metadata associated to the DNARecords as a dictionary of names to pandas DataFrames.
        :param vkeys_columns: columns to return from variant metadata files (potentially big files). Defaults to None (all columns).
        :param skeys_columns: columns to return from sample metadata files (potentially big files). Defaults to None (all columns).
        :param taste: The full metadata DataFrames could be huge, wo you can get a taste of them without going into memory issues. With that, decide wich columns to get metadata for. Defaults to False.

        See Also
        --------
        :obj:`.DNARecordsUtils.dnarecords_tree` to know the available keys
        """
        import dnarecords as dr

        result = {}
        tree = dr.helper.DNARecordsUtils.dnarecords_tree(self._dnarecords_path)
        for k, v in tree.items():
            if k == 'skeys':
                result.update({k: self._pandas_safe_read_parquet(v, skeys_columns, taste)})
            if k == 'vkeys':
                result.update({k: self._pandas_safe_read_parquet(v, vkeys_columns, taste)})
            if k in ['swpfs', 'vwpfs', 'swrfs', 'vwrfs']:
                result.update({k: self._pandas_safe_read_parquet(v, None, False)})
            if k in ['swpsc', 'vwpsc', 'swrsc', 'vwrsc']:
                result.update({k: self._pandas_safe_read_json(v)})
        return result

    def datafiles(self) -> Dict[str, List[str]]:
        """Gets the paths of the DNARecords dataset files as a dictionary of names to List of paths.

        :rtype: Dict[str, List[str]].
        :return: the dnarecord files associated to the DNARecords as a dictionary of names to List of paths.

        See Also
        --------
        :obj:`.DNARecordsUtils.dnarecords_tree` to know the available keys
        """
        import tensorflow as tf
        import dnarecords as dr

        result = {}
        tree = dr.helper.DNARecordsUtils.dnarecords_tree(self._dnarecords_path)
        result.update({'swrec': list(set(tf.io.gfile.glob(f"{tree['swrec']}/*.tfrecord") +
                                         tf.io.gfile.glob(f"{tree['swrec']}/*.tfrecord.gz")))})
        result.update({'vwrec': list(set(tf.io.gfile.glob(f"{tree['vwrec']}/*.tfrecord") +
                                         tf.io.gfile.glob(f"{tree['vwrec']}/*.tfrecord.gz")))})
        result.update({'swpar': list(set(tf.io.gfile.glob(f"{tree['swpar']}/*.parquet") +
                                         tf.io.gfile.glob(f"{tree['swpar']}/*.parquet.gz")))})
        result.update({'vwpar': list(set(tf.io.gfile.glob(f"{tree['vwpar']}/*.parquet") +
                                         tf.io.gfile.glob(f"{tree['vwpar']}/*.parquet.gz")))})
        return result

    @staticmethod
    def _sw_decoder(dnarecords, schema, gzip):
        import json
        import tensorflow as tf
        one_proto = next(iter(tf.data.TFRecordDataset(dnarecords, 'GZIP' if gzip else None)))
        swrsc_dict = {f['fields']['name']: f['fields'] for _, f in schema.iterrows()}
        features = {'key': tf.io.FixedLenFeature([], tf.int64)}
        for indices_field in [field for field in swrsc_dict.keys() if field.endswith('indices')]:
            feature_name = indices_field.replace('_indices', '')
            dense_shape_field = indices_field.replace('indices', 'dense_shape')
            values_field = indices_field.replace('indices', 'values')
            values_type = DNARecordsReader._types_dict()[json.loads(swrsc_dict[values_field]['type'])['elementType']]
            dense_shape = tf.io.parse_example(one_proto, {dense_shape_field: tf.io.FixedLenFeature([], tf.int64)})[
                dense_shape_field]
            features.update({feature_name: tf.io.SparseFeature(indices_field, values_field, values_type, dense_shape)})
        return lambda proto: tf.io.parse_example(proto, features)

    @staticmethod
    def _vw_decoder(dnarecords, schema, gzip):
        import json
        import tensorflow as tf

        one_proto = next(iter(tf.data.TFRecordDataset(dnarecords, 'GZIP' if gzip else None)))
        vwrsc_dict = {f['fields']['name']: f['fields'] for _, f in schema.iterrows()}
        values_type = DNARecordsReader._types_dict()[json.loads(vwrsc_dict['values']['type'])['elementType']]
        dense_shape = tf.io.parse_example(one_proto, {'dense_shape': tf.io.FixedLenFeature([], tf.int64)})[
            'dense_shape']
        features = {'key': tf.io.FixedLenFeature([], tf.int64),
                    'tensor': tf.io.SparseFeature('indices', 'values', values_type, dense_shape)}
        return lambda proto: tf.io.parse_example(proto, features)

    # pylint: disable=too-many-arguments
    # It is reasonable in this case.
    def _dataset(self, dnarecords, decoder, num_parallel_reads, num_parallel_calls, deterministic, drop_remainder,
                 batch_size, buffer_size):
        import tensorflow as tf

        ds = tf.data.TFRecordDataset(dnarecords, 'GZIP' if self._gzip else None, buffer_size, num_parallel_reads)
        if batch_size:
            ds = ds.batch(batch_size, drop_remainder, num_parallel_calls, deterministic)
        return ds.map(decoder, num_parallel_calls, deterministic)

    # pylint: disable=too-many-arguments
    # It is reasonable in this case.
    def sample_wise_dataset(self, num_parallel_reads: int = -1, num_parallel_calls: int = -1,
                            deterministic: bool = False, drop_remainder: bool = False, batch_size: int = None,
                            buffer_size: int = None) -> 'Dataset':
        """DNARecords Tensorflow reader in a sample wise fashion.

        Specially intended for Deep Learning models.

        :return: a Tensorflow dataset with the sample wise DNARecords genomics data.
        :param num_parallel_reads: tf.data.TFRecordDataset equivalent parameter.
        :param num_parallel_calls: tf.data.TFRecordDataset equivalent parameter.
        :param deterministic: tf.data.TFRecordDataset equivalent parameter.
        :param drop_remainder: tf.data.TFRecordDataset equivalent parameter.
        :param batch_size: tf.data.TFRecordDataset equivalent parameter.
        :param buffer_size: tf.data.TFRecordDataset equivalent parameter.
        :rtype: tf.data.Dataset.
        """
        dnarecords = self.datafiles()['swrec']
        schema = self.metadata()['swrsc']
        if schema is None or not dnarecords:
            raise Exception(f"No DNARecords found at {self._dnarecords_path}/...")
        decoder = self._sw_decoder(dnarecords, schema, self._gzip)
        return self._dataset(dnarecords, decoder, num_parallel_reads, num_parallel_calls, deterministic, drop_remainder,
                             batch_size, buffer_size)

    # pylint: disable=too-many-arguments
    # It is reasonable in this case.
    def variant_wise_dataset(self, num_parallel_reads: int = -1, num_parallel_calls: int = -1,
                             deterministic: bool = False, drop_remainder: bool = False, batch_size: int = None,
                             buffer_size: int = None) -> 'Dataset':
        """DNARecords Tensorflow reader in a variant wise fashion.

        Specially intended for GWAS analysis.

        :param num_parallel_reads: tf.data.TFRecordDataset equivalent parameter.
        :param num_parallel_calls: tf.data.TFRecordDataset equivalent parameter.
        :param deterministic: tf.data.TFRecordDataset equivalent parameter.
        :param drop_remainder: tf.data.TFRecordDataset equivalent parameter.
        :param batch_size: tf.data.TFRecordDataset equivalent parameter.
        :param buffer_size: tf.data.TFRecordDataset equivalent parameter.
        :return: a Tensorflow dataset with the variant wise DNARecords genomics data.
        :rtype: tf.data.Dataset.
        """
        dnarecords = self.datafiles()['vwrec']
        schema = self.metadata()['vwrsc']
        if schema is None or not dnarecords:
            raise Exception(f"No DNARecords found at {self._dnarecords_path}/...")
        decoder = self._vw_decoder(dnarecords, schema, self._gzip)
        return self._dataset(dnarecords, decoder, num_parallel_reads, num_parallel_calls, deterministic, drop_remainder,
                             batch_size, buffer_size)


class DNASparkReader:
    """DNARecords Spark reader. **Sample and variant wise**.

    Provides methods to read a previously created DNARecords dataset as pyspark DataFrame objects.

    Review :obj:`.DNARecordsUtils.dnarecords_tree` to know the structure of a previously created DNARecords dataset.

    For each of these directories (when they exist, depending on the configuration used at
    :obj:`.DNARecordsWriter.write`), a spark DataFrame can be returned.


    Example
    --------

    .. code-block:: python

        import dnarecords as dr


        output = '/tmp/dnarecords/output'
        reader = dr.reader.DNASparkReader(output)

        reader.sample_wise_dnarecords().show(2)

        reader.variant_wise_dnarecords().show(2)

    .. code-block:: text

        +---+--------------------+--------------------+----------------+
        |key|        chr1_indices|         chr1_values|chr1_dense_shape| ...
        +---+--------------------+--------------------+----------------+
        | 26|[0, 2, 4, 5, 6, 7...|[0.33607214002352...|             909| ...
        | 29|[0, 1, 2, 3, 4, 5...|[0.20076008098505...|             909| ...
        +---+--------------------+--------------------+----------------+
        only showing top 1 row

        +--------------------+--------------------+----+-----------+
        |             indices|              values| key|dense_shape|
        +--------------------+--------------------+----+-----------+
        |[0, 1, 2, 3, 4, 5...|[0.9984177, 0.007...|3506|      10880|
        |[0, 1, 2, 3, 4, 5...|[0.11181577, 0.01...|3764|      10880|
        +--------------------+--------------------+----+-----------+
        only showing top 2 rows

        ...

    :rtype: Dict[str, DataFrame]
    :param dnarecords_path: path to the DNARecords root directory.
    :return: A dictionary with DataFrame values corresponding to the generated outputs.

    See Also
    --------
    :obj:`.DNARecordsUtils.dnarecords_tree`
    """

    from typing import TYPE_CHECKING
    from typing import Dict

    if TYPE_CHECKING:
        from pyspark.sql import DataFrame

    def __init__(self, dnarecords_path):
        self._dnarecords_path = dnarecords_path

    @staticmethod
    def _spark_safe_load(reader, path):
        try:
            return reader.load(path)
        except Exception as ex:
            raise Exception(f"No DNARecords found at {path}/...") from ex

    def metadata(self) -> Dict[str, 'DataFrame']:
        """Gets the metadata associated to the DNARecords dataset as a dictionary of names to pyspark.sql DataFrames.

        :rtype: Dict[str, DataFrame].
        :return: the metadata associated to the DNARecords as a dictionary of names to pyspark.sql DataFrames.

        See Also
        --------
        :obj:`.DNARecordsUtils.dnarecords_tree` to know the available keys
        """
        import dnarecords as dr

        result = {}
        tree = dr.helper.DNARecordsUtils.dnarecords_tree(self._dnarecords_path)
        spark = dr.helper.DNARecordsUtils.spark_session()
        for k, v in tree.items():
            if k in ['skeys', 'vkeys', 'swpfs', 'vwpfs', 'swrfs', 'vwrfs']:
                result.update({k: self._spark_safe_load(spark.read.format("parquet"), v)})
            if k in ['swpsc', 'vwpsc', 'swrsc', 'vwrsc']:
                result.update({k: self._spark_safe_load(spark.read.format("json"), v)})
        return result

    def sample_wise_dnarecords(self) -> 'DataFrame':
        """Gets a pyspark Dataframe from sample wise DNARecords (when created as tfrecords).

        :rtype: DataFrame.
        :return: a pyspark Dataframe from sample wise DNARecords.
        """
        import dnarecords as dr

        spark = dr.helper.DNARecordsUtils.spark_session()
        path = dr.helper.DNARecordsUtils.dnarecords_tree(self._dnarecords_path)['swrec']
        return self._spark_safe_load(spark.read.format("tfrecord").option("recordType", "Example"), path)

    def variant_wise_dnarecords(self) -> 'DataFrame':
        """Gets a pyspark Dataframe from variant wise DNARecords (when created as tfrecords).

        :rtype: DataFrame.
        :return: a pyspark Dataframe from variant wise DNARecords.
        """
        import dnarecords as dr

        spark = dr.helper.DNARecordsUtils.spark_session()
        path = dr.helper.DNARecordsUtils.dnarecords_tree(self._dnarecords_path)['vwrec']
        return self._spark_safe_load(spark.read.format("tfrecord").option("recordType", "Example"), path)

    def sample_wise_dnaparquet(self) -> 'DataFrame':
        """Gets a pyspark Dataframe from sample wise DNARecords (when created as parquet).

        :rtype: DataFrame.
        :return: a pyspark Dataframe from sample wise DNARecords.
        """
        import dnarecords as dr

        spark = dr.helper.DNARecordsUtils.spark_session()
        path = dr.helper.DNARecordsUtils.dnarecords_tree(self._dnarecords_path)['swpar']
        return self._spark_safe_load(spark.read.format("parquet"), path)

    def variant_wise_dnaparquet(self) -> 'DataFrame':
        """Gets a pyspark Dataframe from variant wise DNARecords (when created as parquet).

        :rtype: DataFrame.
        :return: a pyspark Dataframe from variant wise DNARecords.
        """
        import dnarecords as dr

        spark = dr.helper.DNARecordsUtils.spark_session()
        path = dr.helper.DNARecordsUtils.dnarecords_tree(self._dnarecords_path)['vwpar']
        return self._spark_safe_load(spark.read.format("parquet"), path)



"""
DNARecords available macros.
"""


# pylint: disable=too-few-public-methods
# It is reasonable in this case.
class DosageSparsityMacro:
    """Pre-created methods to increase sparsity of your datasets"""
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from hail import MatrixTable

    @staticmethod
    def supercharge_dosage_sparsity(mt: 'MatrixTable', info_score_threshold: float = 0.8,
                                    p_value_hwe_threshold: float = 1e-10, af_threshold: float = 0.001,
                                    sparse_threshold: float = 0.1) -> 'MatrixTable':
        """ Supercharges dosage sparsity based on info_score and variant_qc.

        Assumes the input MatrixTable to have info_score and variant_qc row fields, and dosage entry field.

        Note
        ----
        It flips the dosage (2 - dosage) for those variants where alt allele is more frequent
        and includes a flag indicating whether the flip was done or not.

        Example
        --------

        .. code-block:: python

            import dnarecords as dr

            hl = dr.helper.DNARecordsUtils.init_hail()
            hl.utils.get_1kg('/tmp/1kg')
            mt = hl.read_matrix_table('/tmp/1kg/1kg.mt')
            mt = mt.annotate_rows(info_score=hl.agg.info_score(hl.pl_to_gp(mt.PL)))
            mt = hl.variant_qc(mt)
            mt = mt.annotate_entries(dosage=hl.pl_dosage(mt.PL))
            mt = dr.macros.DosageSparsityMacro.supercharge_dosage_sparsity(mt)
            mt.select_cols().select_rows('dosage_flip') \\
               .select_entries('dosage','sparse_dosage').entries().show()

        .. code-block:: text

            +---------------+------------+-------------+-----------+----------+---------------+
            | locus         | alleles    | dosage_flip | s         |   dosage | sparse_dosage |
            +---------------+------------+-------------+-----------+----------+---------------+
            | locus<GRCh37> | array<str> |        bool | str       |  float64 |       float64 |
            +---------------+------------+-------------+-----------+----------+---------------+
            | 1:904165      | ["G","A"]  |       False | "HG00096" | 5,94e-02 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00099" | 3,97e-03 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00105" | 4,99e-03 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00118" | 7,88e-03 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00129" | 3,07e-02 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00148" | 7,36e-02 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00177" | 2,01e-01 |      2,01e-01 |
            | 1:904165      | ["G","A"]  |       False | "HG00182" | 3,83e-02 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00242" | 3,07e-02 |      0,00e+00 |
            | 1:904165      | ["G","A"]  |       False | "HG00254" | 1,26e-04 |      0,00e+00 |
            +---------------+------------+-------------+-----------+----------+---------------+

        :param mt: a MatrixTable with **info_score** and **variant_qc** row fields, and **dosage** entry field.
        :param info_score_threshold: rows with info_score.score below the threshold are filtered out.
        :param p_value_hwe_threshold: rows with variant_qc.p_value_hwe below the threshold are filtered out.
        :param af_threshold: rows with AF < af_threshold or AF > (1 - af_threshold) are filtered out.
        :param sparse_threshold: those entries with dosage below the threshold are set dosage = 0.
        :return: a MatrixTable with above transformations done.
        :rtype: MatrixTable
        """
        from dnarecords.helper import DNARecordsUtils
        hl = DNARecordsUtils.init_hail()
        mt = mt.filter_rows(info_score_threshold < mt.info_score.score)
        mt = mt.filter_rows(p_value_hwe_threshold < mt.variant_qc.p_value_hwe)
        mt = mt.filter_rows((af_threshold < mt.variant_qc.AF[1]) & (mt.variant_qc.AF[1] < (1 - af_threshold)))
        mt = mt.annotate_entries(sparse_dosage=hl.if_else(0.5 < mt.variant_qc.AF[1], 2 - mt.dosage, mt.dosage))
        mt = mt.annotate_rows(dosage_flip=(0.5 < mt.variant_qc.AF[1]))
        mt = mt.annotate_entries(sparse_dosage=hl.if_else(mt.sparse_dosage < sparse_threshold, 0.0, mt.sparse_dosage))
        return mt



"""
DNARecords available writers.
"""


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
# It is reasonable in this case.
class DNARecordsWriter:
    """Genomics data (vcf, bgen, etc.) to tfrecords or parquet. **Sample and variant wise**.

    Core class to go from genomics data to tfrecords or parquet files ready to use with
    Deep Learning frameworks like Tensorflow or Pytorch.

    * Able to generate **dna records variant wise or sample wise (i.e. transposing the matrix)**.
    * **Takes advantage of sparsity** (very convenient to save space and computation, specially with Deep Learning).
    * **Scales automatically to any sized dataset. Tested on UKBB**.

    Example
    --------

    .. code-block:: python

        import dnarecords as dr


        hl = dr.helper.DNARecordsUtils.init_hail()
        hl.utils.get_1kg('/tmp/1kg')
        mt = hl.read_matrix_table('/tmp/1kg/1kg.mt')
        mt = mt.annotate_entries(dosage=hl.pl_dosage(mt.PL))

        output = '/tmp/dnarecords/output'
        writer = dr.writer.DNARecordsWriter(mt.dosage)
        writer.write(output, sparse=True, sample_wise=True, variant_wise=True,
                     tfrecord_format=True, parquet_format=True,
                     write_mode='overwrite', gzip=True)

        reader = dr.reader.DNASparkReader(output)

        reader.sample_wise_dnarecords().show(2)

        reader.variant_wise_dnarecords().show(2)

    .. code-block:: text

        +---+--------------------+--------------------+----------------+
        |key|        chr1_indices|         chr1_values|chr1_dense_shape| ...
        +---+--------------------+--------------------+----------------+
        | 26|[0, 2, 4, 5, 6, 7...|[0.33607214002352...|             909| ...
        | 29|[0, 1, 2, 3, 4, 5...|[0.20076008098505...|             909| ...
        +---+--------------------+--------------------+----------------+
        only showing top 1 row

        +--------------------+--------------------+----+-----------+
        |             indices|              values| key|dense_shape|
        +--------------------+--------------------+----+-----------+
        |[0, 1, 2, 3, 4, 5...|[0.9984177, 0.007...|3506|      10880|
        |[0, 1, 2, 3, 4, 5...|[0.11181577, 0.01...|3764|      10880|
        +--------------------+--------------------+----+-----------+
        only showing top 2 rows

        ...

    :param expr: a Hail expression. Currently, ony expressions coercible to numeric are supported
    :param block_size: entries per block in the internal operations
    :param staging: path to staging directory to use for intermediate data. Default: /tmp/dnarecords/staging.
    """
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from hail import MatrixTable, Expression
        from pyspark.sql import DataFrame

    _i_blocks: set
    _j_blocks: set
    _nrows: int
    _ncols: int
    _sparsity: float
    _chrom_ranges: dict
    _mt: 'MatrixTable'
    _skeys: 'DataFrame'
    _vkeys: 'DataFrame'

    def __init__(self, expr: 'Expression', block_size: int = int(1e6), staging: str = '/tmp/dnarecords/staging'):
        self._assert_expr_type(expr)
        self._expr = expr
        self._block_size = block_size
        self._kv_blocks_path = f'{staging}/kv-blocks'
        self._vw_dna_staging = f'{staging}/vw-dnaparquet'
        self._sw_dna_staging = f'{staging}/sw-dnaparquet'

    @staticmethod
    def _assert_expr_type(expr):
        import hail.expr.expressions.expression_typecheck as tc
        msg = 'Only expr_numeric are supported at this time.'
        msg += '\nConsider redefining the entry_expr as an expr_numeric.'
        msg += '\nFeel free to use the helper functions on this module.'
        if not tc.expr_numeric.can_coerce(expr.dtype):
            raise Exception(msg)

    def _set_mt(self):
        from hail.expr.expressions.expression_utils import matrix_table_source
        self._mt = matrix_table_source('dnarecords', self._expr).annotate_entries(v=self._expr)

    def _index_mt(self):
        self._mt = self._mt.add_row_index('i')
        self._mt = self._mt.key_rows_by('i')
        self._mt = self._mt.add_col_index('j')
        self._mt = self._mt.key_cols_by('j')

    def _set_vkeys_skeys(self):
        self._vkeys = self._mt.key_rows_by().rows().to_spark().withColumnRenamed('i', 'key').cache()
        self._skeys = self._mt.key_cols_by().cols().to_spark().withColumnRenamed('j', 'key').cache()

    def _set_chrom_ranges(self):
        from pyspark.sql import functions as F
        gdf = self._vkeys.select('`locus.contig`', 'key').groupby('`locus.contig`')
        gdf = gdf.agg(F.min('key').alias('start'), F.max('key').alias('end'))
        self._chrom_ranges = {r['locus.contig'].replace('chr', ''): [r['start'], r['end']] for i, r in
                              gdf.toPandas().iterrows()}

    def _update_vkeys_by_chrom_ranges(self):
        from dnarecords.helper import DNARecordsUtils
        import pyspark.sql.functions as F
        chr_start_keys = [[k, start] for k, [start, _] in self._chrom_ranges.items()]
        spark = DNARecordsUtils.spark_session()
        chr_start_keys_df = spark.sparkContext.parallelize(chr_start_keys).toDF(['locus.contig', 'chr_start_key'])
        self._vkeys = self._vkeys \
            .join(chr_start_keys_df, on='locus.contig', how='left') \
            .withColumn('chr_key', F.col('key') - F.col('chr_start_key'))

    def _select_ijv(self):
        self._mt = self._mt.select_globals().select_rows().select_cols().select_entries('v')

    def _filter_out_undefined_entries(self):
        from dnarecords.helper import DNARecordsUtils
        hl = DNARecordsUtils.init_hail()
        self._mt = self._mt.filter_entries(hl.is_defined(self._mt.v))

    def _filter_out_zeroes(self):
        from dnarecords.helper import DNARecordsUtils
        hl = DNARecordsUtils.init_hail()
        self._mt = self._mt.filter_entries(0 != hl.coalesce(self._mt.v, 0))

    def _set_max_nrows_ncols(self):
        self._nrows = self._mt.count_rows()
        self._ncols = self._mt.count_cols()

    def _set_sparsity(self):
        mts = self._mt.head(10000, None)
        entries = mts.key_cols_by().key_rows_by().entries().to_spark().filter('v is not null').count()
        self._sparsity = entries / (mts.count_rows() * mts.count_cols())

    def _get_block_size(self):
        import math
        M, N, S = self._nrows + 1, self._ncols + 1, self._sparsity + 1e-6
        B = self._block_size / S
        m = math.ceil(math.sqrt(B * M / N))
        n = math.ceil(math.sqrt(B * N / M))
        return m, n

    def _build_ij_blocks(self):
        import pyspark.sql.functions as F
        m, n = self._get_block_size()
        df = self._mt.key_cols_by().key_rows_by().entries().to_spark().filter('v is not null')
        df = df.withColumn('ib', F.expr(f"i div {m}"))
        df = df.withColumn('jb', F.expr(f"j div {n}"))
        df.repartition('ib', 'jb').write.partitionBy('ib', 'jb').mode('overwrite').parquet(self._kv_blocks_path)

    def _set_ij_blocks(self):
        import re
        from dnarecords.helper import DNARecordsUtils
        hl = DNARecordsUtils.init_hail()
        all_blocks = [p for p in hl.hadoop_ls(f'{self._kv_blocks_path}/*') if p['is_dir']]
        self._i_blocks = {re.search(r'ib=(\d+)', p['path']).group(1) for p in all_blocks}
        self._j_blocks = {re.search(r'jb=(\d+)', p['path']).group(1) for p in all_blocks}

    @staticmethod
    def _dnarecord_variant_agg(rows, mkey):
        for row in rows:
            sd = dict(sorted(row['data'].items()))
            yield {'key': row['i'], 'indices': list(sd.keys()), 'values': list(sd.values()),
                   'dense_shape': mkey + 1}

    @staticmethod
    def _dnarecord_sample_agg(rows, chrom_ranges):
        import numpy as np
        for row in rows:
            sd = dict(sorted(row['data'].items()))
            keys = np.array(list(sd.keys()))
            vals = np.array(list(sd.values()))
            result = {'key': row['j']}
            for locus, [start, end] in chrom_ranges.items():
                mask = (start <= keys) & (keys <= end)
                result.update({f'chr{locus}_indices': (keys[mask] - start).tolist(),
                               f'chr{locus}_values': vals[mask].tolist(),
                               f'chr{locus}_dense_shape': end - start + 1})
            yield result

    @staticmethod
    def _to_dnarecord_variant_wise(mkey):
        return lambda rows: DNARecordsWriter._dnarecord_variant_agg(rows, mkey)

    @staticmethod
    def _to_dnarecord_sample_wise(chrom_ranges):
        return lambda rows: DNARecordsWriter._dnarecord_sample_agg(rows, chrom_ranges)

    def _build_dna_block_variant_wise(self, blocks_path, output):
        from dnarecords.helper import DNARecordsUtils
        import pyspark.sql.functions as F

        def get_dna_schema(gdf):
            import pyspark.sql.types as pytypes

            data_type = [s for s in gdf.schema if s.name == 'data'][0].dataType
            values_type = data_type.valueType
            return pytypes.StructType([pytypes.StructField("key", pytypes.LongType(), False),
                                       pytypes.StructField("indices", pytypes.ArrayType(pytypes.LongType(), False),
                                                           False),
                                       pytypes.StructField("values", pytypes.ArrayType(values_type, False), False),
                                       pytypes.StructField("dense_shape", pytypes.LongType(), False)])

        spark = DNARecordsUtils.spark_session()
        df = spark.read.parquet(blocks_path).select('i', 'j', 'v')
        df = df.groupBy('i').agg(F.map_from_entries(F.collect_list(F.struct('j', 'v'))).alias('data'))
        schema = get_dna_schema(df)
        mapper = self._to_dnarecord_variant_wise(self._nrows)
        df.rdd.mapPartitions(mapper).toDF(schema).repartition(1).write.mode('overwrite').parquet(output)

    def _build_dna_block_sample_wise(self, blocks_path, output, chrom_ranges):
        from dnarecords.helper import DNARecordsUtils
        import pyspark.sql.functions as F

        def get_dna_schema(gdf):
            import pyspark.sql.types as pytypes

            data_type = [s for s in gdf.schema if s.name == 'data'][0].dataType
            values_type = data_type.valueType
            gdf_schema = pytypes.StructType([pytypes.StructField("key", pytypes.LongType(), False)])
            for k in chrom_ranges.keys():
                gdf_schema.add(f"chr{k}_indices", pytypes.ArrayType(pytypes.LongType(), False), False)
                gdf_schema.add(f"chr{k}_values", pytypes.ArrayType(values_type, False), False)
                gdf_schema.add(f"chr{k}_dense_shape", pytypes.LongType(), False)
            return gdf_schema

        spark = DNARecordsUtils.spark_session()
        df = spark.read.parquet(blocks_path).select('i', 'j', 'v')
        df = df.groupBy('j').agg(F.map_from_entries(F.collect_list(F.struct('i', 'v'))).alias('data'))
        schema = get_dna_schema(df)
        mapper = self._to_dnarecord_sample_wise(chrom_ranges)
        df.rdd.mapPartitions(mapper).toDF(schema).repartition(1).write.mode('overwrite').parquet(output)

    def _build_dna_blocks(self, by):
        from multiprocessing.pool import ThreadPool
        import multiprocessing as ms

        pool = ThreadPool(ms.cpu_count() - 1)
        if by == 'i':
            params = [
                [f'{self._kv_blocks_path}/ib={ib}/jb=*', f'{self._vw_dna_staging}/{ib}-of-{len(self._i_blocks):04}']
                for ib in self._i_blocks]
            pool.starmap(self._build_dna_block_variant_wise, params)
        if by == 'j':
            params = [
                [f'{self._kv_blocks_path}/ib=*/jb={jb}', f'{self._sw_dna_staging}/{jb}-of-{len(self._j_blocks):04}',
                 self._chrom_ranges]
                for jb in self._j_blocks]
            pool.starmap(self._build_dna_block_sample_wise, params)
        pool.close()
        pool.join()

    # pylint: disable=too-many-arguments
    # It is reasonable in this case.
    @staticmethod
    def _write_dnarecords(output, output_schema, dna_blocks, write_mode, gzip, tfrecord_format):
        from dnarecords.helper import DNARecordsUtils

        spark = DNARecordsUtils.spark_session()
        df = spark.read.parquet(dna_blocks)
        df_writer = df.write.mode(write_mode)
        if tfrecord_format:
            df_writer = df_writer.format("tfrecord").option("recordType", "Example")
            if gzip:
                # Needs huge overhead memory
                df_writer = df_writer.option("codec", "org.apache.hadoop.io.compress.GzipCodec")
        else:
            df_writer = df_writer.format('parquet')
            if gzip:
                df_writer = df_writer.option("compression", "gzip")
        df_writer.save(output)
        sc_writer = spark.read.json(spark.sparkContext.parallelize([df.schema.json()])).coalesce(1).write
        sc_writer.mode(write_mode).format('json').save(output_schema)

    @staticmethod
    def _write_key_files(source, output, tfrecord_format, write_mode):
        from dnarecords.helper import DNARecordsUtils
        import pyspark.sql.functions as F

        spark = DNARecordsUtils.spark_session()
        if tfrecord_format:
            reader = spark.read.format("tfrecord").option("recordType", "Example")
        else:
            reader = spark.read.format("parquet")
        df = reader.load(source).withColumn("path", F.regexp_extract(F.input_file_name(), f"(.*){source}/(.*)", 2))
        df.select('key', 'path').write.mode(write_mode).parquet(output)

    # pylint: disable=too-many-arguments
    # It is reasonable in this case.
    def write(self, output: str, sparse: bool = True, sample_wise: bool = True, variant_wise: bool = False,
              tfrecord_format: bool = True, parquet_format: bool = False, write_mode: str = 'error',
              gzip: bool = True) -> None:
        """DNARecords spark writer.

        Writes a DNARecords dataset based on the Hail `expr` provided in the class constructor.

        :rtype: Dict[str, DataFrame]
        :param output: path to the output location of the DNARecords.
        :param sparse: generate sparse data (filtering out any zero values). Default: True.
        :param sample_wise: generate DNARecords in a sample wise fashion (i.e. transposing the matrix, one column -> one record). Default: True.
        :param variant_wise: generate DNARecords in a variant wise fashion (i.e. one row -> one record) Default: False.
        :param tfrecord_format: generate tfrecords output files. Default: True.
        :param parquet_format: generate parquet output files. Default: False.
        :param write_mode: spark write mode parameter ('error', 'overwrite', etc.). Default: 'error'.
        :param gzip: gzip the output files. Default: True.
        :return: A dictionary with DataFrames for each generated output.

        See Also
        --------
        :obj:`.DNARecordsUtils.dnarecords_tree`, :obj:`.DNASparkReader`
        """
        from dnarecords.helper import DNARecordsUtils

        if not sample_wise and not variant_wise:
            raise Exception('At least one of sample_wise, variant_wise must be True')
        if not tfrecord_format and not parquet_format:
            raise Exception('At least one of tfrecord_format, parquet_format must be True')

        otree = DNARecordsUtils.dnarecords_tree(output)

        self._set_mt()
        self._index_mt()
        self._set_vkeys_skeys()
        self._set_chrom_ranges()
        self._update_vkeys_by_chrom_ranges()

        self._vkeys.write.mode(write_mode).parquet(otree['vkeys'])
        self._skeys.write.mode(write_mode).parquet(otree['skeys'])

        self._select_ijv()
        self._filter_out_undefined_entries()
        if sparse:
            self._filter_out_zeroes()
        self._set_max_nrows_ncols()
        self._set_sparsity()
        self._build_ij_blocks()
        self._set_ij_blocks()

        if variant_wise:
            self._build_dna_blocks('i')

        if sample_wise:
            self._build_dna_blocks('j')

        if variant_wise:
            if tfrecord_format:
                self._write_dnarecords(otree['vwrec'], otree['vwrsc'], f'{self._vw_dna_staging}/*', write_mode, gzip,
                                       True)
                self._write_key_files(otree['vwrec'], otree['vwrfs'], True, write_mode)
            if parquet_format:
                self._write_dnarecords(otree['vwpar'], otree['vwpsc'], f'{self._vw_dna_staging}/*', write_mode, gzip,
                                       False)
                self._write_key_files(otree['vwpar'], otree['vwpfs'], False, write_mode)

        if sample_wise:
            if tfrecord_format:
                self._write_dnarecords(otree['swrec'], otree['swrsc'], f'{self._sw_dna_staging}/*', write_mode,
                                       gzip, True)
                self._write_key_files(otree['swrec'], otree['swrfs'], True, write_mode)
            if parquet_format:
                self._write_dnarecords(otree['swpar'], otree['swpsc'], f'{self._sw_dna_staging}/*', write_mode,
                                       gzip, False)
                self._write_key_files(otree['swpar'], otree['swpfs'], False, write_mode)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dnarecords']

package_data = \
{'': ['*']}

install_requires = \
['fastparquet>=0.8.1,<0.9.0',
 'hail>=0.2.94,<0.3.0',
 'pyarrow>=8.0.0,<9.0.0',
 'tensorflow>=2.6.0,<3.0.0']

setup_kwargs = {
    'name': 'dnarecords',
    'version': '0.2.3',
    'description': 'Genomics data ML ready',
    'long_description': "# DNARecords\n\n[![PyPI license](https://img.shields.io/pypi/l/ansicolortags.svg)](https://pypi.python.org/pypi/ansicolortags/)\n![example workflow](https://github.com/amanas/dnarecords/actions/workflows/ci-cd.yml/badge.svg)\n[![codecov](https://codecov.io/gh/amanas/dnarecords/branch/main/graph/badge.svg)](https://codecov.io/gh/amanas/dnarecords)\n![pylint Score](https://mperlet.github.io/pybadge/badges/9.97.svg)\n[![semantic-release: angular](https://img.shields.io/badge/semantic--release-angular-e10079?logo=semantic-release)](https://github.com/semantic-release/semantic-release)\n\n**Genomics data ML ready.**\n\nTransform your vcf, bgen, etc. genomics datasets into a sample wise format so that you can use it \nfor Deep Learning models. \n\n## Installation\n\nDNARecords package has two main dependencies:\n\n* **Hail**, if you are transforming your genomics data into DNARecords\n* **Tensorflow**, if you are using a previously DNARecords dataset, for example, to train a DL model\n\nAs you may know, Tensorflow and Spark does not play very well together on a cluster with more than one machine.\n\nHowever, `dnarecords` package needs to be installed **only on the driver machine** of a Hail cluster.\n\nFor that reason, we recommend following these installation tips.\n\n### **On a dev environment**\n\n```bash\n$ pip install dnarecords\n```\n\nFor further details (or any trouble), review [Local environments](LOCAL_ENVS.md) section.\n\n### **On a Hail cluster or submitting a job to it**\n\nYou will already have Pyspark installed and will not intend to install Tensorflow. \n\nSo, just install dnarecords without dependencies on the driver machine. \n\nThere will be no missing modules as soon as you use the classes and functionality intended for Spark.\n\n```bash\n$ /opt/conda/miniconda3/bin/python -m pip install dnarecords --no-deps\n```\n*Note: assuming Hail python executable is /opt/conda/miniconda3/bin/python* \n\n### **On a Tensorflow environment or submitting a job to it**\n\nYou will already have Tensorflow installed and will not intend to install Pyspark. \n\nSo, just install dnarecords without dependencies. \n\nThere will be no missing modules as soon as you use the classes and functionality intended for Tensorflow.\n\n```bash\n$ pip install dnarecords --no-deps\n```\n\n\n## Working on Google Dataproc\n\nJust use and initialization action that installs `dnarecords` without dependencies.\n\n```bash\n$ hailctl dataproc start dnarecords --init gs://dnarecords/dataproc-init.sh\n```\nIy you need to work with other cloud providers, refer to [Hail docs](https://hail.is/docs/0.2/getting_started.html).\n\n## Usage\n\nIt is quite straightforward to understand the functionality of the package.\n\nGiven some genomics data, you can transform it into a **DNARecords Dataset** this way:\n\n\n```python\nimport dnarecords as dr\n\n\nhl = dr.helper.DNARecordsUtils.init_hail()\nhl.utils.get_1kg('/tmp/1kg')\nmt = hl.read_matrix_table('/tmp/1kg/1kg.mt')\nmt = mt.annotate_entries(dosage=hl.pl_dosage(mt.PL))\n\ndnarecords_path = '/tmp/dnarecords'\nwriter = dr.writer.DNARecordsWriter(mt.dosage)\nwriter.write(dnarecords_path, sparse=True, sample_wise=True, variant_wise=True,\n             tfrecord_format=True, parquet_format=True,\n             write_mode='overwrite', gzip=True)\n```\n\n\nGiven a DNARecords Dataset, you can read it as **Tensorflow Datasets** this way:\n\n```python\nimport dnarecords as dr\n\n\ndnarecords_path = '/tmp/dnarecords'\nreader = dr.reader.DNARecordsReader(dnarecords_path)\nsamplewise_ds = reader.sample_wise_dataset()\nvariantwise_ds = reader.variant_wise_dataset()\n```\n\nOr, given a DNARecords Dataset, you can read it as **Pyspark DataFrames** this way:\n\n\n```python\nimport dnarecords as dr\n\n\ndnarecords_path = '/tmp/dnarecords'\nreader = dr.reader.DNASparkReader(dnarecords_path)\nsamplewise_df = reader.sample_wise_dnarecords()\nvariantwise_df = reader.variant_wise_dnarecords()\n```\n\nWe will provide more examples and integrations soon.\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`dnarecords` was created by Atray Dixit, Andrés Mañas Mañas, Lucas Seninge. It is licensed under the terms of the MIT license.\n\n## Credits\n\n`dnarecords` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n",
    'author': 'Atray Dixit, Andrés Mañas Mañas, Lucas Seninge',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)
